/*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.

* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
* Copyright (C) 2009 - Marcelo Leal <msl@eall.com.br>. All rights reserved.
* Copyright (C) 2004 - iMil <imil@gcu.info>. All rights reserverd.
*/

#include <thread.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <menu.h>
#include <curses.h>

#define  _GNU_SOURCE
#define  NBLOCKS 9
#define  TAM512 512 
#define _REENTRANT
#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))
#define CTRLD 4
#define _CONF_INTEGER   1
#define _CONF_STRING    2
#define CONF_INTEGER(x, y)              { x, &y, _CONF_INTEGER, 0 }
#define CONF_NULL                       { NULL, NULL, 0, 0 }
#define CONF_STRPTR_INTEGER(w, z)       CONF_INTEGER(#z, w->z)
#define LINHAS 19
#define COLUNAS 80
#define SUBCOL 40 
#define BORDER 9
#define BOTTOM 5
#define DEFAULT_SCALE 100
#define DEFAULT_DELAY 1
#define WATTRON(a, b) if (colors != 0) { \
        wattron(a, b); \
}
#define WATTROFF(a, b) if (colors != 0) { \
        wattroff(a, b); \
}
#define MAX(a,b) ((a)<(b)?(b):(a))
#define ABS(a) ((a)>0?(a):-(a))
#define INTERVAL (5)

/* Array data structures */

struct {
        mutex_t data_lock[10];
        int     rvalues[10];
        int     wvalues[10];
        int     trvalues[10];
        int     twvalues[10];
        } data;

char *blocksizes[] = {
			"A       <=    1KB ",
			"B >  1KB <=   2KB ",
			"C >  2KB <=   4KB ",
			"D >  4KB <=   8KB ",
			"E >  8KB <=  16KB ",
			"F > 16KB <=  32KB ",
			"G > 32KB <=  64KB ",
			"H > 64KB <= 128KB ",
                        "I > 128KB         ",
			(char *)NULL,
		};

static void finish(int);
static void init_window_top();
static void init_windows_graphs();
static void *read_data(void *);
static void draw_marks(WINDOW *, int);
static void draw_numbers( int, int, int);
static void draw_graph(int, int *, int, WINDOW *, int *);
static int *update_array(int, int *, int *);
void print_str_middle(WINDOW *, int, int, int, char *, chtype);
void print_int_middle(WINDOW *, int, int, int, int *, chtype, int);

//Global variables
//I know, i know, but it's easier...
const char *letters[9]={"A","B","C","D","E","F","G","H","I"};
int rpercent[10], wpercent[10]; 
int r_max, w_max, mega=0, totalr=0, totalw=0, *pmega;
int a, i, j, y, l, z=10, *rgraph = NULL, *wgraph = NULL; 
int bdtr[10], *pbdtr, bdtw[10], *pbdtw, elapsedtr[10], elapsedtw[10], *peltr, *peltw;
int bdr[10], *pbdr, bdw[10], *pbdw, elapsedr[10], elapsedw[10], *pelr, *pelw;
int window_h, colors;
MENU *menu;
ITEM **itens_menu;
WINDOW *window_pri, *window_sec, *window_ter;

static void
finish(int sig)
{
        delwin(window_pri);
        delwin(window_sec);
        endwin();

        free(rgraph);
        free(wgraph);

        exit(0);
}

static void
init_window_top()
{
 // Criacao do Menu principal 
 menu = new_menu((ITEM **)itens_menu);

 // Nao mostrar a descricao de cada opcao...
 menu_opts_off(menu, O_SHOWDESC);

 // Criacao da janela a ser associada com o menu 
 window_pri  = newwin(LINHAS - 3, COLUNAS, 1, 4);
 window_sec  = newwin(LINHAS, SUBCOL, 17, 4);
 window_ter  = newwin(LINHAS, SUBCOL, 17, 44);
 nodelay (window_pri, TRUE);
 nodelay (window_sec, TRUE);
 nodelay (window_ter, TRUE);
 leaveok (window_pri, TRUE);
 leaveok (window_sec, TRUE);
 leaveok (window_ter, TRUE);
 window_h = LINHAS + 2 - BOTTOM;

 // Janela principal e sub menu.
 set_menu_win(menu, window_pri);
 set_menu_sub(menu, derwin(window_pri, 11, 76, 5, 1));
 set_menu_format(menu, 17, 1);

 // Cursor 
 set_menu_mark(menu, " ");

 // Borda e titulo do menu 
 box(window_pri, 0, 0);
 print_str_middle(window_pri, 1, 20, 40, "---=| NMT (v0.5) |=---", COLOR_PAIR(0));
 print_str_middle(window_pri, 3, 9, 40, "R", COLOR_PAIR(0));
 print_str_middle(window_pri, 3, 16, 40, "%", COLOR_PAIR(0));
 print_str_middle(window_pri, 3, 24, 40, "TR", COLOR_PAIR(0));
 print_str_middle(window_pri, 3, 37, 40, "W", COLOR_PAIR(0));
 print_str_middle(window_pri, 3, 45, 40, "%", COLOR_PAIR(0));
 print_str_middle(window_pri, 3, 54, 40, "TW", COLOR_PAIR(0));

 // Menu 
 post_menu(menu);

 mvwaddch(window_pri, 2, 0, ACS_LTEE);
 mvwhline(window_pri, 2, 1, ACS_HLINE, 79);
 mvwaddch(window_pri, 2, 79, ACS_RTEE);

 mvwaddch(window_pri, 4, 0, ACS_LTEE);
 mvwhline(window_pri, 4, 1, ACS_HLINE, 79);
 mvwaddch(window_pri, 4, 79, ACS_RTEE);
 wrefresh(window_pri);

}

static void
init_windows_graphs() {
 box(window_sec, 0, 0);
 box(window_ter, 0, 0);
 mvwaddch(window_sec, 0, 35, 'R');
 mvwaddch(window_sec, 0, 36, 'E');
 mvwaddch(window_sec, 0, 37, 'A');
 mvwaddch(window_sec, 0, 38, 'D');

 mvwaddch(window_ter, 0, 34, 'W');
 mvwaddch(window_ter, 0, 35, 'R');
 mvwaddch(window_ter, 0, 36, 'I');
 mvwaddch(window_ter, 0, 37, 'T');
 mvwaddch(window_ter, 0, 38, 'E');

 draw_marks(window_sec, 0);
 draw_marks(window_ter, 0);

}

static void
draw_marks (WINDOW *win, int color)
{
 //AXIS
 mvwhline(win, 16, 12, ACS_HLINE, SUBCOL -30);
 mvwvline(win, 1, 11, ACS_VLINE, 15);
 mvwaddch(win, 1, 11, ACS_ULCORNER);
 mvwaddch(win, 16, 11, ACS_LLCORNER);

 mvwhline(win, 16, 25, ACS_HLINE, SUBCOL -28);
 mvwvline(win, 7, 24, ACS_VLINE, 9);
 mvwaddch(win, 6, 24, ACS_ULCORNER);
 mvwaddch(win, 16, 24, ACS_LLCORNER);
 mvwhline(win, 6, 25, ACS_HLINE, SUBCOL -28);
 mvwvline(win, 7, 37, ACS_VLINE, 9);
 mvwaddch(win, 6, 37, ACS_URCORNER);
 mvwaddch(win, 16, 37, ACS_LRCORNER);

 mvwaddch(win, 6, 28, 'r');
 mvwaddch(win, 6, 29, 'e');
 mvwaddch(win, 6, 30, 'q');
 mvwaddch(win, 6, 31, '/');
 mvwaddch(win, 6, 32, 's');

 WATTRON(win, COLOR_PAIR(color));
 wattron(win, A_BOLD);
 for (i = 0; i < NBLOCKS; i++) {
     mvwaddch(win, 17, i + BORDER + 4, *letters[i]);
     mvwaddch(win, i + 7, 26, *letters[i]);
 }
 WATTROFF(win, COLOR_PAIR(color));
 wattroff(win, A_BOLD);
 //wrefresh(win);

}

static int *
update_array(int num, int *array, int *max)
{
        int i;
        *max = 0;

        if (array == NULL) {
                /* we malloc NBLOCKS elements */
                array = (int *)malloc(NBLOCKS * sizeof(int));
                for (i = 0; i < NBLOCKS; i++)
                        array[i] = 0;
        } else
                for (i = NBLOCKS-1; i >=0; i--) {
                        array[i] = array[i-1];
                        if (*max < array[i])
                                *max = array[i];
                }

        array[0] = num;
        return (array);
}

static void
draw_graph(int scale, int *array, int color, WINDOW *win, int *gp)
{
   int a, i, j, nbar, nscale = 0;
 
   /* draw grid and scale */
   for (i = BOTTOM; i <= window_h + 1; i += 2) {
       nscale++;
   }

   nscale = scale / nscale;
   WATTRON(win, COLOR_PAIR(color));
   for (i = 0, j = 0; i < ( window_h - 1 ); i += 2, j++) {
       mvwprintw(win, window_h - i, 1, " %dMB/s ", j * nscale);
       mvwaddch(win, window_h - i, 12, ACS_HLINE);
   }
   WATTROFF(win, COLOR_PAIR(color));

   /* draw graph */
   WATTRON(win, COLOR_PAIR(color));
   wattron(win, A_BOLD);

   for (i = 0; i < NBLOCKS; i++) {
       nbar = (array[i] * window_h) / (scale - 4);
       if (nbar > 15) {
          nbar = 15;
       }
       for (a=0;a<nbar;a++){
           mvwaddch(win, window_h - a -1, i + BORDER + 4, ACS_VLINE);
       }
       mvwprintw(win, i + 7, 30, "%d", gp[i]);
   }
   WATTROFF(win, COLOR_PAIR(color));
   wattroff(win, A_BOLD);
}

void print_str_middle(WINDOW *win, int y, int x, int width, char *string, chtype color) {
   int length;
   float temp;
   if(width == 0)
   width = 80;
   length = strlen(string);
   temp = (width - length)/ 2;
   x = x + (int)temp;
   wattron(win, color);
   mvwprintw(win, y, x, "%s", string);
   wattroff(win, color);
   wnoutrefresh(win);
   doupdate();
}

void print_int_middle(WINDOW *win, int y, int x, int width, int *inteiro, chtype color, int format) {
   wattron(win, color);
   if (format > 1) {
      mvwprintw(win, x, y, "%4d", *inteiro);
   } else {
      mvwprintw(win, x, y, "%8lu", *inteiro);
   }
   wattroff(win, color);
   wnoutrefresh(win);
   doupdate();
}

// Read data...
static void 
*read_data(void *nulo) 
{
    struct dtraceoutput {
        int r;
        int w;
    } dtracedata;
    struct dtraceoutput *pdtrace = &dtracedata;

    struct blockdbfields {
        char *op_name;
        int *op_variable;
        int op_type;
        int op_maxlength;
    };

    struct blockdbfields
           blockdb[] = {
             CONF_STRPTR_INTEGER(pdtrace, r),
             CONF_STRPTR_INTEGER(pdtrace, w),
             CONF_NULL,
    };
    struct blockdbfields *p;
    // We have this in some place, but i don't know where... ;)
    int v1k=1024, v2k=2048, v4k=4096, v8k=8192; 
    int v16k=16384, v32k=32768, v64k=65536, v128k=131072;

    int *p1=&v1k, *p2=&v2k, *p3=&v4k, *p4=&v8k;
    int *p5=&v16k, *p6=&v32k, *p7=&v64k, *p8=&v128k;

    FILE *f;

    char buffer[TAM512];
    char *vname, *end, *value, *temp=NULL, *errorptr;
    int i, intvalue, ln, length, oper=0;

    ln = 0;
    while (fgets(buffer, sizeof(buffer), stdin)) {
        ln++;
        length = strlen(buffer);
        if (buffer[length - 1] != '\n') {
            while (fgets(buffer, sizeof(buffer), stdin)) {
                if (buffer[strlen(buffer) - 1] == '\n') break;
            }
            continue;
        }
        buffer[length - 1] = '\0';
        for (vname = buffer; *vname == ' ' || *vname == '\t'; vname++);
        if (*vname == '#' || *vname == '\0') continue;
        if (!(value = strchr(vname, '='))) {
            exit(1);
            continue;
        }
        for (end = value - 1; *end == ' ' || *end == '\t'; end--);
        end[1] = '\0';
        for (p = blockdb; p->op_name; p++) {
            if (strcmp(p->op_name, vname) == 0) break;
        }
        if (!p->op_name) {
            printf("Operation unknown: %s\n", vname);
            exit(1);
        }
        for (value++; *value == ' ' || *value == '\t'; value++);
        for (end = buffer + length - 2; *end == ' ' || *end == '\t'; end--);
        end++; *end = '\0';

        // Always an integer...
        intvalue = strtol(value, &errorptr, 0);
        if (*errorptr != '\0') {
             printf("Invalid: %s\n", vname);
             exit(1);
         }
        *((int *)p->op_variable) = intvalue;


        //Read or Write...
        if (temp = strchr(p->op_name, 'r')) {
         oper = 1;
        } else {
         oper = 2;
        }

        // Draw...
        if ((*p->op_variable > *p1) && (*p->op_variable <= *p2))
        {
           draw_numbers (1, oper, *p->op_variable);
         } else if ((*p->op_variable > *p2) && (*p->op_variable <= *p3)) 
           {
           draw_numbers (2, oper, *p->op_variable);
         } else if ((*p->op_variable > *p3) && (*p->op_variable <= *p4)) 
           {
           draw_numbers (3, oper, *p->op_variable);
         } else if ((*p->op_variable > *p4) && (*p->op_variable <= *p5)) 
           {
           draw_numbers (4, oper, *p->op_variable);
           } else if ((*p->op_variable > *p5) && (*p->op_variable <= *p6)) 
         {
           draw_numbers (5, oper, *p->op_variable);
         } else if ((*p->op_variable > *p6) && (*p->op_variable <= *p7))
           {
           draw_numbers (6, oper, *p->op_variable);
         } else if ((*p->op_variable > *p7) && (*p->op_variable <= *p8))
           {
           draw_numbers (7, oper, *p->op_variable);
         } else if (*p->op_variable > *p8)
           {
           draw_numbers (8, oper, *p->op_variable);
         } else {
           draw_numbers (0, oper, *p->op_variable);
        }
    }
    exit(0);
}

static void
draw_numbers( int line, int oper, int p)
{
  int i, coord=0, *gp, *gpt, *gpp, *pt=NULL;
  int *pwpercent = wpercent, *prpercent = rpercent;
  WINDOW *win = window_pri;

  //Read or Write...
  if (oper == 1) {
   coord = 21;
   gp  = data.rvalues;
   gpt = data.trvalues;
   gpp = prpercent;
   pt  = &totalr;
  } else {
   coord= 50;
   gp  = data.wvalues;
   gpt = data.twvalues;
   gpp = pwpercent;
   pt  = &totalw;
  }
  mutex_lock(&data.data_lock[line]);
       gp[line] = gp[line] + 1;
       print_int_middle(win, coord,      line+5, 5, &gp[line], COLOR_PAIR(0), 1);
       gpt[line] = (gpt[line] + p);
       print_int_middle(win, coord + 18, line+5, 5, &gpt[line], COLOR_PAIR(0), 1);
       *pt = ((*pt) + 1);
       gpp[line] = ((100 * gp[line]) / *pt);
       print_int_middle(win, coord + 11, line+5, 5, &gpp[line], COLOR_PAIR(0), 2);
  mutex_unlock(&data.data_lock[line]);
}

// ----------------= The main function =-----------------
int main (int argc, const char **argv) {
 int scaler = DEFAULT_SCALE, scalew = DEFAULT_SCALE;
 int new_scaler=0, new_scalew=0;
 int autoscale = 1; pmega = &mega;
 pbdr = bdr; pbdw = bdw;
 pbdtr = bdtr; pbdtw = bdtw;
 pelr = elapsedr; pelw = elapsedw;
 peltr = elapsedtr; peltw = elapsedtw;
 void *nulo;
 int *pr = rpercent, *pw = wpercent;

 /* initialize the mutexes and data */
 for (i=0; i<NBLOCKS; i++) {
     mutex_init(&data.data_lock[i], USYNC_THREAD, 0);
     data.rvalues[i] = 0;
     data.wvalues[i] = 0;
     data.trvalues[i] = 0;
     data.twvalues[i] = 0;
     pr[i] = 0; 
     pw[i] = 0;
 }

 /* set concurrency and create the threads */
 thr_setconcurrency(2);

 // Blocks and values array
 itens_menu = (ITEM **)calloc(NBLOCKS, sizeof(ITEM *));

 for (i = 0; i < NBLOCKS; i++)
 {
   itens_menu[i] = new_item(blocksizes[i], blocksizes[i]);
 }

 signal(SIGINT, finish);
 // Formatacao da tela... inicio do ncurses. 
 initscr();
 if ((scaler - LINHAS) < 0)
                scaler = LINHAS;
 if ((scalew - LINHAS) < 0)
                scalew = LINHAS;

 if (has_colors() == FALSE)
        colors = 0;
    else {
        colors = 1;
        start_color();
        init_pair(1, COLOR_GREEN, -1);
        init_pair(2, COLOR_RED, -1);
        init_pair(3, COLOR_BLUE, -1);
        init_pair(4, COLOR_WHITE, -1);
        init_pair(5, COLOR_RED, -1);
        init_pair(6, COLOR_CYAN, -1);
        init_pair(7, COLOR_MAGENTA, -1);
        init_pair(8, COLOR_WHITE, -1);
 }
 //assume_default_colors(-1,-1);
 nonl();
 cbreak();
 noecho();
 attron(COLOR_PAIR(0));
 mvprintw(LINES - 5, 30, "Pressione: <Ctrl C> para SAIR");
 mvprintw(LINES - 3, 29, "- http://www.eALL.com.br/blog -");
 mvprintw(LINES - 2, 20, "NFS Read/Write Monitoring Tool (Powered by Dtrace)");
 attroff(COLOR_PAIR(0));
 refresh();
 if ((LINES < 41 ) || (COLS < 85)) {
    mvprintw(LINES - 1, 20, "Sorry, but you need a terminal at least 41x85 wide.");
    refresh();
    exit(1);
 }

 init_window_top();
 init_windows_graphs();
 
 //Go, and don't come back!
 thr_create(NULL, 0, read_data, (void *)nulo, 0, NULL);

 //Wait, so we have some data...
 sleep(1);

 //Our turn...
 while (1) {
       for (i=0;i<NBLOCKS;i++){
           mutex_lock(&data.data_lock[i]);
                pbdr[i] = data.rvalues[i]; pbdw[i] = data.wvalues[i];
                pbdtr[i] = data.trvalues[i]; pbdtw[i] = data.twvalues[i];
           mutex_unlock(&data.data_lock[i]);
       }
       sleep(1);
       for (i=0;i<NBLOCKS;i++){
           mutex_lock(&data.data_lock[i]);
                pelr[i] = data.rvalues[i]; pelw[i] = data.wvalues[i];
                peltr[i] = data.trvalues[i]; peltw[i] = data.twvalues[i];
           mutex_unlock(&data.data_lock[i]);
       }
       for (i=NBLOCKS-1;i>=0;i--){
           *pmega = ((peltr[i] - pbdtr[i]) / (1024*1024));
           rgraph = update_array(*pmega, rgraph, &r_max);
           *pmega = ((peltw[i] - pbdtw[i]) / (1024*1024));
           wgraph = update_array(*pmega, wgraph, &w_max);
           pbdr[i] = (pelr[i] - pbdr[i]);
           pbdw[i] = (pelw[i] - pbdw[i]);
       }
       if (autoscale != 0) {
          new_scaler = r_max + INTERVAL;
          new_scalew = w_max + INTERVAL;
          if (ABS(new_scaler - scaler) >= INTERVAL) {
             scaler = new_scaler;
             if ((scaler / LINHAS) < 1) {
                scaler = LINHAS;
             }
          }
          if (ABS(new_scalew - scalew) >= INTERVAL) {
             scalew = new_scalew;
             if ((scalew / LINHAS) < 1) {
                scalew = LINHAS;
             }
          }
       }
       wclear(window_sec);
       wclear(window_ter);
       init_windows_graphs();
       draw_graph(scaler, rgraph, 0, window_sec, pbdr);
       wnoutrefresh(window_sec);
       draw_graph(scalew, wgraph, 0, window_ter, pbdw);
       wnoutrefresh(window_ter);
       sleep(DEFAULT_DELAY);
 }
} /* End of the main function */
